package Comp;
import AST.*;
import java.util.*;
import Lexer.*;
import java.io.*;


public class Compiler {
    
      // compile must receive an input with an character less than 
      // p_input.lenght
   public Program compile( char []input, PrintWriter outError ) {
        
       symbolTable = new SymbolTable();
       error = new CompilerError( lexer, new PrintWriter(outError) );
       lexer = new Lexer(input, error);
       error.setLexer(lexer);
        
        
       Program p = null;
       try {
          lexer.nextToken();
          if ( lexer.token == Symbol.EOF )
             error.show("Unexpected EOF");
          p = program();
          if ( lexer.token != Symbol.EOF ) {
             p = null;
             error.show("EOF expected");
          }
       } catch ( Exception e ) {
              // the below statement prints the stack of called methods.
              // of course, it should be removed if the compiler were 
              // a production compiler.
              //e.printStackTrace();
          p = null;
       }
       
       return p;   
   }
    
   
   private Program program() {
      classDec();
      while ( lexer.token == Symbol.CLASS ) 
         classDec();
      return null;
   }

   private void classDec() {
      
      if ( lexer.token != Symbol.CLASS ) 
         error.show("'class' expected");
      lexer.nextToken();
      if ( lexer.token != Symbol.IDENT )
         error.show(CompilerError.identifier_expected);
      String className = lexer.getStringValue();
      
      lexer.nextToken();
      if ( lexer.token == Symbol.EXTENDS ) {
         lexer.nextToken();
         if ( lexer.token != Symbol.IDENT )
            error.show(CompilerError.identifier_expected);
         String superclassName = lexer.getStringValue();
         
         lexer.nextToken();
      }
      if ( lexer.token != Symbol.LEFTCURBRACKET )
         error.show("{ expected", true);
      lexer.nextToken();
      
      
      while ( lexer.token == Symbol.PRIVATE ||
              lexer.token == Symbol.PUBLIC  ||
              lexer.token == Symbol.FINAL ) {
         
         int qualifier;
         switch( lexer.token ) {
            case Symbol.PRIVATE:
               lexer.nextToken();
               qualifier = Symbol.PRIVATE;
               break;
            case Symbol.FINAL :
               lexer.nextToken();
               if ( lexer.token != Symbol.PUBLIC )
                  error.show("'public' expected");
               lexer.nextToken();
               qualifier = Symbol.FINAL;
               break;
            case Symbol.PUBLIC:
               lexer.nextToken();
               qualifier = Symbol.PUBLIC;
               break;
            default :
               error.show("private, final, or public expected");
               qualifier = Symbol.PUBLIC;
         }
         Type t = type();
         if ( lexer.token != Symbol.IDENT )
            error.show("Identifier expected");
         String name = lexer.getStringValue();
         lexer.nextToken();
         if ( lexer.token == Symbol.LEFTPAR )
            methodDec(t, name, qualifier);
         else if ( qualifier != Symbol.PRIVATE )
            error.show("Attempt to declare a public instance variable");
         else
            instanceVarDec(t, name);
      }
      if ( lexer.token != Symbol.RIGHTCURBRACKET )
         error.show("public/private/final or \"}\" expected");
      lexer.nextToken();
      
   }
   
   
   private void instanceVarDec( Type type, String name ) {
    
      while ( lexer.token == Symbol.COMMA ) {
         lexer.nextToken();
         if ( lexer.token != Symbol.IDENT )
            error.show("Identifier expected");
         String variableName = lexer.getStringValue();
         lexer.nextToken();
      }
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }
   
   

   private void methodDec(Type type, String name, int qualifier ) {
    
      lexer.nextToken();
      if ( lexer.token != Symbol.RIGHTPAR )
         formalParamDec();
      if ( lexer.token != Symbol.RIGHTPAR )
         error.show(") expected");      
      
      lexer.nextToken();
      if ( lexer.token != Symbol.LEFTCURBRACKET ) 
         error.show("{ expected");
         
      lexer.nextToken();
      statementList();
      if ( lexer.token != Symbol.RIGHTCURBRACKET ) 
         error.show("} expected");
         
      lexer.nextToken();
         
   }




   private void localDec( Type type ) {
    
            
      if ( lexer.token != Symbol.IDENT ) 
         error.show("Identifier expected");
      Variable v = new Variable(lexer.getStringValue(), type);
      lexer.nextToken();
      while ( lexer.token == Symbol.COMMA ) {
         lexer.nextToken();
         if ( lexer.token != Symbol.IDENT ) 
            error.show("Identifier expected");
         v = new Variable(lexer.getStringValue(), type);
         lexer.nextToken();
      }
   }

   
   
          
   private void formalParamDec() {

      paramDec();
      while ( lexer.token == Symbol.COMMA ) {
         lexer.nextToken();
         paramDec();
      }
   }

   private void paramDec() {

      type();
      if ( lexer.token != Symbol.IDENT )
         error.show("Identifier expected");
      lexer.nextToken();
   }
        

    private Type type() {
        Type result;
        
        switch ( lexer.token ) {
            case Symbol.INT :
               result = Type.intType;
               break;
            case Symbol.BOOLEAN :
               result = Type.booleanType;
               break;
            case Symbol.STRING :
               result = Type.stringType;
               break;
            case Symbol.VOID :
               result = Type.voidType;
               break;
            case Symbol.IDENT :
               //# corrija: faa uma busca na TS para buscar a classe
               // IDENT deve ser uma classe.
               result = null;
               break;
            default :
              error.show("Type expected");
              result = Type.undefinedType;
        }
        lexer.nextToken();
        return result;
    }
    
        
        

    private void compositeStatement() {
        
        lexer.nextToken();
        statementList();
        if ( lexer.token != Symbol.RIGHTCURBRACKET )
          error.show("} expected");
        else
          lexer.nextToken();
    }



   private void statementList() {
       int tk;
          // statements always begin with an identifier, if, read, write, ...
       while ( (tk = lexer.token) != Symbol.RIGHTCURBRACKET && 
               tk != Symbol.ELSE ) 
          statement();
   }
            
   private void statement() {
           
      switch (lexer.token) {
         case Symbol.IDENT :
         case Symbol.SUPER :
         case Symbol.THIS :
         case Symbol.INT :
         case Symbol.BOOLEAN :
         case Symbol.STRING:
            assignmentMessageSendLocalVarDecStatement();
            break;
         case Symbol.IF :
            ifStatement();
            break;
         case Symbol.WHILE :
            whileStatement();
            break;
         case Symbol.RETURN :
            returnStatement();
            break;
         case Symbol.READ :
            readStatement();
            break;
         case Symbol.WRITE :
            writeStatement();
            break;
         case Symbol.BREAK :
            breakStatement();
            break;
         case Symbol.SEMICOLON :
            nullStatement();
            break;
         case Symbol.LEFTCURBRACKET :
            compositeStatement();
            break;
         default :
            error.show("Statement expected");
      }
   }
    
   private Statement assignmentMessageSendLocalVarDecStatement() {
      // an assignment, a message send or a local variable declaration
    
      String methodName, variableName;
      ExprList exprList;
      Statement result = null;
      /*
         there are eight possibilities:
            super.id()
            this.id()
            this.id = expr
            this.id.id()
            id = expr
            id.id()
            
            Id IdList  // Id is a type
            int IdList or boolean IdList or String IdList
      */
      
      switch ( lexer.token ) {
         case Symbol.SUPER :
               // super.id()
            lexer.nextToken();
            if ( lexer.token != Symbol.DOT ) 
               error.show(". expected");            
            lexer.nextToken();
            if ( lexer.token != Symbol.IDENT )
               error.show(CompilerError.identifier_expected);
            methodName = lexer.getStringValue();
            lexer.nextToken();
            exprList = getRealParameters();
            //# corrija
            // result = new MessageSendStatement( 
            //     new MessageSendToSuper( pointer to class, pointer to method, exprList) );
            break;
         case Symbol.THIS :
            lexer.nextToken();
            if ( lexer.token != Symbol.DOT ) 
               error.show(". expected");
            lexer.nextToken();
            if ( lexer.token != Symbol.IDENT )
               error.show(CompilerError.identifier_expected);
            String ident = lexer.getStringValue();
            lexer.nextToken();
            switch ( lexer.token ) {
               case Symbol.ASSIGN :
                     // this.id = expr 
                  lexer.nextToken();
                  Expr anExpr = expr();
                  //# corrija
                  /* result = new AssignmentStatement( pointer to instance variable,
                        anExpr); */
                  break;
               case Symbol.DOT :
                     // this.id.id() 
                  lexer.nextToken();
                  if ( lexer.token != Symbol.IDENT )
                     error.show(CompilerError.identifier_expected);
                  methodName = lexer.getStringValue();
                  lexer.nextToken();
                  exprList = getRealParameters();
                  //# corrija
                  /* result = new MessageSendStatement( 
                        new MessageSendToVariable( pointer to variable,
                           pointer to method, exprList) );  */
                  break;
               case Symbol.LEFTPAR :
                     // this.id()
                  exprList = getRealParameters();
                  //# corrija
                  /* result = new MessageSendStatement( 
                       new MessageSendToSelf( pointer to method, exprList ) ); */
                  break;
               default :
                  error.show(CompilerError.identifier_expected);
            }
            break;
         case Symbol.IDENT :
            variableName = lexer.getStringValue();
            lexer.nextToken();
            switch ( lexer.token ) {
               case Symbol.IDENT :
                    // id id;
                    // variableName id
                    
                    // variableName must be the name of a class
                    // replace null in the statement below by 
                    // a point to the class named variableName.
                    // A search in the symbol table is necessary.
                  localDec(null);
                  break;
               case Symbol.ASSIGN : 
                     // id = expr
                  lexer.nextToken();
                  Expr anExpr = expr();
                  //# corrija
                  /* result = new AssignmentStatement( pointer to variable,
                        anExpr ); */
                  break;
               case Symbol.DOT :
                     // id.id()
                  lexer.nextToken();
                  methodName = lexer.getStringValue();
                  lexer.nextToken();
                  exprList = getRealParameters();
                  //# corrija
                  /* result = new MessageSendStatement( 
                         new MessageSendToVariable( pointer to variable,
                               pointer to method, exprList ) ); */
                  break;
               default :
                  error.show(". or = expected");
            }
            break;
         case Symbol.INT :
            lexer.nextToken();
            localDec(Type.intType);
            break;
         case Symbol.BOOLEAN :
            lexer.nextToken();
            localDec(Type.booleanType);
            break;
         case Symbol.STRING:
            lexer.nextToken();
            localDec(Type.stringType);
            break;
         default :
            error.show("'this', 'super', basic type or identifier expected");
      }
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
      return result;
   }

    
   private ExprList getRealParameters() {
      ExprList anExprList = null;
      
      if ( lexer.token != Symbol.LEFTPAR ) 
         error.show("( expected");
      lexer.nextToken();
      if ( startExpr(lexer.token) )
         anExprList = exprList();
      if ( lexer.token != Symbol.RIGHTPAR )
         error.show(") expected");
      lexer.nextToken();
      return anExprList;
   }
    
   private void ifStatement() {
        
      lexer.nextToken();
      if ( lexer.token != Symbol.LEFTPAR ) 
         error.show("( expected");
      lexer.nextToken();
      expr();
      if ( lexer.token != Symbol.RIGHTPAR ) 
         error.show(") expected");
      lexer.nextToken();
      statement();
      if ( lexer.token == Symbol.ELSE ) {
         lexer.nextToken();
         statement();
      }
   }
            
   private void whileStatement() {

      lexer.nextToken();
      if ( lexer.token != Symbol.LEFTPAR ) 
         error.show("( expected");
      lexer.nextToken();
      expr();
      if ( lexer.token != Symbol.RIGHTPAR ) 
         error.show(") expected");
      lexer.nextToken();
      statement();
   }
    
    
   private void returnStatement() {
        
      lexer.nextToken();
      expr();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }
          
            
   private void readStatement() {
      lexer.nextToken();
      if ( lexer.token != Symbol.LEFTPAR ) 
         error.show("( expected");
      lexer.nextToken();
      while ( true ) {
         if ( lexer.token == Symbol.THIS ) {
            lexer.nextToken();
            if ( lexer.token != Symbol.DOT ) 
               error.show(". expected");
            lexer.nextToken();
         }
         if ( lexer.token != Symbol.IDENT )
            error.show(CompilerError.identifier_expected);
         
         String name = (String ) lexer.getStringValue();
         lexer.nextToken();
         if ( lexer.token == Symbol.COMMA )
            lexer.nextToken();
         else
            break;
      }
      
      if ( lexer.token != Symbol.RIGHTPAR ) 
         error.show(") expected");
      lexer.nextToken();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }
    
    
   private void writeStatement() {
        
      lexer.nextToken();
      if ( lexer.token != Symbol.LEFTPAR ) 
         error.show("( expected");
      lexer.nextToken();
      exprList();
      if ( lexer.token != Symbol.RIGHTPAR ) 
          error.show(") expected");
      lexer.nextToken();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }
        

   private void breakStatement() {
      lexer.nextToken();
      if ( lexer.token != Symbol.SEMICOLON )
         error.show(CompilerError.semicolon_expected);
      lexer.nextToken();
   }

   
   private void nullStatement() {
      lexer.nextToken();
   }
   
    
   private ExprList exprList() {
    
      ExprList anExprList = new ExprList();
      anExprList.addElement(expr());
      while ( lexer.token == Symbol.COMMA ) {
         lexer.nextToken();
         anExprList.addElement(expr());
      }
      return anExprList;
   }



   private Expr expr() {
    
      Expr left = simpleExpr();
      int op = lexer.token;
      if ( op == Symbol.EQ || op == Symbol.NEQ || 
           op == Symbol.LE || op == Symbol.LT ||
           op == Symbol.GE || op == Symbol.GT  ) {
         lexer.nextToken();
         Expr right = simpleExpr();
         left = new CompositeExpr( left, op, right );
      }
      return left;
   }


   private Expr simpleExpr() {
      int op;
      
      Expr left = term();
      while ( (op = lexer.token) == Symbol.PLUS || op == Symbol.MINUS || 
              op == Symbol.OR ) {
         lexer.nextToken();
         Expr right = term();
         left = new CompositeExpr( left, op, right );
      }
      return left;
   }


   private Expr term() {
      int op;
      
      Expr left = signalFactor();
      while ( (op = lexer.token) == Symbol.MULT || op == Symbol.DIV || 
              op == Symbol.AND ) {
         lexer.nextToken();
         Expr right = signalFactor();
         left = new CompositeExpr( left, op, right );
      }
      return left;
   }
   
   
   private Expr signalFactor() {
      int op;
      if ( (op = lexer.token) == Symbol.PLUS || op == Symbol.MINUS ) {
         lexer.nextToken();
         return new SignalExpr(op, factor());
      }
      else
         return factor();
   }
   


    private Expr factor() {
      /*
        Factor ::= Number | "true" | "false" | "this" [ "." Id ]
           | Id | MessageSend | '(' Expr ')' | "!" Factor | "null"
           | "new" Id "(" ")"
      */
      
      Expr e;
      
      switch ( lexer.token ) {
         case Symbol.NUMBER :
            return number();
         case Symbol.TRUE :
           lexer.nextToken();
           return BooleanExpr.True;
         case Symbol.FALSE :
           lexer.nextToken();
           return BooleanExpr.False;
         case Symbol.LITERALSTRING :
           String literalString = lexer.getLiteralStringValue();
           lexer.nextToken();
           return new LiteralStringExpr(literalString);
         case Symbol.LEFTPAR :
            lexer.nextToken();
            e = expr();
            if ( lexer.token != Symbol.RIGHTPAR ) 
               error.show(") expected");
            lexer.nextToken();
            return new ParenthesisExpr(e);
         case Symbol.NOT :
            lexer.nextToken();
            e = expr();
            return new UnaryExpr( e, Symbol.NOT );
         case Symbol.NULL :
            lexer.nextToken();
            return new NullExpr();
         case Symbol.NEW :
            lexer.nextToken();
            if ( lexer.token != Symbol.IDENT ) 
               error.show("Identifier expected");
            String className = lexer.getStringValue();
            lexer.nextToken();
            if ( lexer.token != Symbol.LEFTPAR ) 
               error.show("( expected");
            lexer.nextToken();
            if ( lexer.token != Symbol.RIGHTPAR )
               error.show(") expected");
            lexer.nextToken();
            // return an object representing the creation of an object
            return null;
         default :
            String variableName, methodName;   
            ExprList exprList;
               /* there are seven cases to consider :
                     x
                     x.m()
                     super.m()
                     this.x
                     this
                     this.m()
                     this.x.m();
                     
                  in which x is either a variable or 
                  an instance variable and m is a method
               */
            switch ( lexer.token ) {
               case Symbol.IDENT :
                  variableName = lexer.getStringValue();
                  lexer.nextToken();
                  if ( lexer.token != Symbol.DOT ) {
                     // expression of the kind "x"
                     //# corrija
                     //return new VariableExpr(pointer to variable);
                  }
                  else {
                       // expression of the kind "x.m()"
                     lexer.nextToken();
                     switch ( lexer.token ) {
                        case Symbol.IDENT :
                           methodName = lexer.getStringValue();
                           lexer.nextToken();
                           exprList = getRealParameters();
                           //#  corrija
                           /* return new MessageSendToVariable(
                                 pointer to variable, pointer to method, exprList) */
                           break;
                        default :
                           error.show(CompilerError.identifier_expected);
                     }
                  }
                  break;
               case Symbol.SUPER :
                    // expression of the kind "super.m()"
                  lexer.nextToken();
                  if ( lexer.token != Symbol.DOT )
                     error.show(". expected");
                  lexer.nextToken();
                  if ( lexer.token != Symbol.IDENT )
                     error.show(CompilerError.identifier_expected);
                  methodName = lexer.getStringValue();
                  lexer.nextToken();
                  exprList = getRealParameters();
                  //#  corrija
                  /* return new MessageSendToSuper(
                         pointer to class, pointer to method, exprList) */
                  break;
               case Symbol.THIS :
                  lexer.nextToken();
                  if ( lexer.token != Symbol.DOT ) {
                     // expression of the kind "this"
                     //# corrija
                     // return new ThisExpr();
                  }
                  else {
                     lexer.nextToken();
                     if ( lexer.token != Symbol.IDENT )
                        error.show(CompilerError.identifier_expected);
                        // it may be method name or an instance variable
                     String ident = lexer.getStringValue();
                     lexer.nextToken();
                     switch ( lexer.token ) {
                        case Symbol.LEFTPAR : 
                           // expression of the kind "this.m()"
                           //# corrija
                           exprList = getRealParameters();
                           /*
                           return new MessageSendToSelf( pointer to method,
                              exprList );
                           */
                           break;
                        case Symbol.DOT :
                           // expression of the kind "this.x.m()"
                           //# corrija
                           lexer.nextToken();
                           if ( lexer.token != Symbol.IDENT )
                              error.show(CompilerError.identifier_expected);
                           methodName = lexer.getStringValue();
                           lexer.nextToken();
                           exprList = getRealParameters();
                           /*
                           return new MessageSendToVariable( 
                              pointer to instance variable ident,
                              pointer to method,
                              exprList );
                           */
                        default :
                           // expression of the kind "this.x"
                           //# corrija
                           /*
                           return new VariableExpr( pointer to instance 
                              variable ident )
                           */
                     }
                        
                  }
                  break;
               default : 
                  error.show(CompilerError.identifier_expected);
               }
               return null;
         }
            
   }

    
   private NumberExpr number() {
        
      NumberExpr e = null;
        
          // the number value is stored in lexer.getToken().value as an object of Integer.
          // Method intValue returns that value as an value of type int.
      int value = lexer.getNumberValue();
      lexer.nextToken();
      return new NumberExpr( value );
   }
    
    
   private boolean startExpr( int aToken ) {
    
      return lexer.token == Symbol.IDENT || 
             lexer.token == Symbol.NUMBER ||
             lexer.token == Symbol.FALSE ||
             lexer.token == Symbol.TRUE ||
             lexer.token == Symbol.THIS || 
             lexer.token == Symbol.SUPER ||
             lexer.token == Symbol.LEFTPAR ||
             lexer.token == Symbol.NOT ||
             lexer.token == Symbol.NULL ||
             lexer.token == Symbol.LITERALSTRING;
           
   }
    
    
   private SymbolTable symbolTable;
   private Lexer lexer;
   private CompilerError error;
   
}
